# *** PACK_GRID_UTILITIES PY - 20 Jun 2020  16:30:46 - JKNAUTH

"""
pack_grid_utilities provides support functions for pack_grid.pyw.  One
function validates the contents of the file identified by pack_grid.pyw
and prepares that file data for execution by tkinter pack.  Another
function invokes tkinter pack for the selected test case(s) and displays
the results.  See pack_grid_documentation.py for details about the main
data structures.
"""

version = 2.2

from tkinter import *
from tkinter.messagebox import showerror, showwarning
import os.path

# ======================================================================
# Specify label color and font settings for various conditions.

# Background color for each each label, depending on its packing side
bg_TOP    = 'cyan'
bg_LEFT   = 'pink'
bg_RIGHT  = 'orange'
bg_BOTTOM = 'lime'

# Font and foreground color for the first label packed
ft_1 =   ('lucida', 10, 'bold underline italic')
fg_1 = 'brown'

# Font and foreground color for labels packed after the first
ft_a =   ('lucida', 9, 'bold')
fg_a = 'black'

# Font and colors for a CASE ID label
ft_case_ID = ('lucida', 7, 'bold')
bg_case_ID = 'tan'
fg_case_ID = 'black'

# Font and colors for a text label
ft_text = ('lucida', 10)
bg_text = 'yellow'
fg_text = 'black'

# Font and colors for an error cell
ft_error = ('lucida', 13, 'bold')
bg_error = 'red'
fg_error = 'black'


# ======================================================================
# Global variables

gapx = 2   # Gaps between the grid cells to allow the backing frame to
gapy = 2   #    show thru and better separate the cells

fileid = ''            # Fileid passed from pack_grid.pyw to be analyzed
cases = {}             # Dictionary of test cases
ID_order = []          # Keys (test case IDs) from the dictionary
file_rows = 'MAX'      # User-requested row size from test case file
nbr_cases = 0          # Number of test cases to be displayed

totcells = 0           # Runnng count of grid cells processed
row = 0                # Current grid row
col = 0                # Current grid column


# $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
# $$$$$$$$$$$$$$$$$$$$$$$$$$  Analyze File  $$$$$$$$$$$$$$$$$$$$$$$$$$$$
# $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$

# This function is called by pack_grid.pyw to analyze a file.  If the
# file is found to be good, the data in it is used to set up a number of
# variables which are saved and later used when pack_grid.pyw calls
# another function below to display the test case results for this file.

def process_fileid(in_fileid):
   global fileid, cases, ID_order, file_rows

   fileid = in_fileid

   # Open the file containing the test cases
   try:
      infile = open(fileid)
   except:
      (directory, filename) = os.path.split(fileid)
      showerror('ERROR!',
                'Directory  ' + directory + '\n\n' +
                'Filename:  ' + filename + '\n\n' +
                'Unable to open file.')
      return 'Cannot open file'

   # ===================================================================
   # Loop thru the test cases in the test case file to build the
   # dictionary and the ID order list.  Skip over certain lines, e.g.,
   # blank lines.  Process a "rows =" line if present.  Error check test
   # case IDs.  Error check each pack string to see if it has the
   # options, values, and case that pack expects; this also ensures that
   # the string is safe to feed to eval.

   cases = {}              # Initialize the test case dictionary
   cases_info = {}         # Initialize info string dictionary
   ID_order = []           # Initialize the ID order list
   file_rows = 'MAX'       # Use MAX if there is no "rows =" in the file

   for line in infile:
      line = line.strip()
      if line == '': continue                    # Skip blank line
      if line.startswith('* *** ', 0): continue  # Skip JGK timestamp line
      if line.startswith('#'): continue          # Skip comment line

      # Process "rows = nbr" or "rows = max" if present.
      if line.startswith('rows'):
         org_line = line          # Save original line for error message
         fr_splt = line.split('=')
         if fr_splt[0].strip() == 'rows' and len(file_rows) > 1:
            file_rows = fr_splt[1].upper().strip()
            if file_rows == 'MAX': continue
            if file_rows.isdigit(): continue
         (directory, filename) = os.path.split(fileid)
         showerror('ERROR!',
                   'Directory  ' + directory + '\n\n' +
                   'Filename:  ' + filename + '\n\n' +
                   'Invalid "rows =" value in this line:   ' + org_line)
         return 'Invalid "rows ="'

      # Get the ID and validate it.
      split_line = line.split(',')
      ID = split_line[0].strip().rjust(3,'0')
      if not ID.isdigit():                # ID must be an integer string
         (directory, filename) = os.path.split(fileid)
         showerror('ERROR!',
                   'Directory  ' + directory + '\n\n' +
                   'Filename:  ' + filename + '\n\n' +
                   'The ID (first field) in this line is not an integer ' +
                   'or the integer is not followed by a comma:\n\n' + line)
         return 'ID is not an integer'

      # Handle lines with little data.
      if len(split_line) == 1: split_line = [ID, 'IGNORE']
      if len(split_line) == 2 and split_line[1] == '': split_line = [ID, 'IGNORE']

      # Check for a possible "TEXT=" line and report an error if "TEXT="
      # is in the wrong place.
      textline = False
      mixed_data = ','.join(split_line[1:]).strip()  # Restore original non-ID
      upper_data = mixed_data.upper()                #    part of the line
      if upper_data.find('TEXT=') >= 0:
         if not upper_data.startswith('TEXT='):
            (directory, filename) = os.path.split(fileid)
            showerror('ERROR!',
                      'Directory  ' + directory + '\n\n' +
                      'Filename:  ' + filename + '\n\n' +
                      'The "TEXT=" in this line is misplaced.  It must be ' +
                      'the first non-blank characters after the ID and its ' +
                      'comma:\n\n' + line)
            return '"TEXT=" in wrong place'
         else:
            textline = True
            # Standardize case for TEXT= at start of text line and handle
            # null text.
            rest = mixed_data[5:]
            if rest == '': rest = '(no text was supplied)'
            pack_string = rest

      # These are the valid pack options=values (and a special IGNORE word).
      valids = ['side=TOP', 'side=BOTTOM', 'side=LEFT', 'side=RIGHT',
                'fill=NONE', 'fill=X', 'fill=Y', 'fill=BOTH',
                'expand=YES', 'expand=NO',
                'anchor=NW',    'anchor=N',   'anchor=NE',
                'anchor=W',  'anchor=CENTER', 'anchor=E',
                'anchor=SW',    'anchor=S',   'anchor=SE',
                'IGNORE']

      # Check for invalid pack options.  Also protects against bad eval.
      if not textline:
         words = split_line[1:]
         # Allow last option to have a trailing comma (pack allows it).
         if words[-1] == '': words = words[0:-1]
         pack_string = ''
         for word in words:
            word = word.replace(' ','')             # Squeeze out spaces
            if word not in valids:
               (directory, filename) = os.path.split(fileid)
               showerror('ERROR!',
                         'Directory  ' + directory + '\n\n' +
                         'Filename:  ' + filename + '\n\n' +
                         'Invalid pack option (' + word + ') in this line:\n' +
                         '(check spelling, including use of proper case)\n' +
                         '(check for extra commas)\n\n' + line)
               return 'Invalid pack option'
            pack_string = pack_string + word + ','  # Reassemble pack string
         pack_string = pack_string[0:-1]            # Delete final comma

      # Add the new info string to the list of info strings for this ID
      # key in the cases_info dictionary.
      if textline:
         case_info_for_ID = cases_info.get(ID,[])
         case_info_for_ID.append(pack_string)
         cases_info[ID] = case_info_for_ID
      else:
         # Add the new pack string to the list of pack strings for this
         # ID key in the test case dictionary.
         case_packs_for_ID = cases.get(ID,[])
         case_packs_for_ID.append(pack_string)
         cases[ID] = case_packs_for_ID

      # Add the ID to the test case order list.  Only the first
      # occurrence of an ID in the file is added to the list since the
      # dictionary merges all data for a specific ID under one key (the
      # ID), no matter how the file lines are arranged.  The ID_order
      # list just preserves the order of the first occurrence of each
      # ID.  See the example in pack_grid_documentation.py.
      if ID not in ID_order: ID_order.append(ID)


   # -------------------------------------------------------------------
   # The test cases file is no longer needed.  Close it.
   infile.close()

   # Check for a file with no test cases in it.
   if len(ID_order) == 0:
      (directory, filename) = os.path.split(fileid)
      showerror('ERROR!',
                'Directory  ' + directory + '\n\n' +
                'Filename:  ' + filename + '\n\n' +
                'There are no test cases in this file.')
      return 'No test cases in the file'

   # -------------------------------------------------------------------
   # All the file lines have now been processed.  Now for each ID that
   # had any "TEXT=" lines, glue together (with a space between each
   # adjacent pair) all the "TEXT=" data sections for that ID.  Then
   # create a "TEXT=" string with the consolidated data as the "TEXT="
   # data section.  Finally, in the cases dictionary, add that "TEXT="
   # string to the ID's pack info list as the first item in the list.

   for ID in cases_info:
      partials_list = cases_info[ID]
      info_string = ' '.join(partials_list).rstrip()
      pack_info = ['TEXT=' + info_string]
      if ID in cases:
         cases[ID][:0] = pack_info
      else:
         cases[ID] = pack_info

   # ===================================================================
   # The fileid processing has gotten to this point with no errors.  Now
   # return to pack_grid.pyw.

   return ''               # A return of '' means the file had no errors



# $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
# $$$$$$$$$$$$$$$$$$$$$$  Build Results Display  $$$$$$$$$$$$$$$$$$$$$$$
# $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
# ======================================================================
# This function is called from pack_grid.pyw to build the display for
# grid mode or "just one" mode.

def build_display_window(fileid, display_parms):
   global cases, ID_order, file_rows, gapx, gapy, comment_labels

   (case_ID, just_one, reduce_size, abbreviate) = display_parms

   # ===================================================================
   # Trim the ID_order list to contain only the IDs that can be
   # displayed.  For "just one" mode, this is simple.  For grid mode, it
   # requires much calculation.

   # For "just one" mode there is only one case.  Trim ID_order to just
   # the one selected case.
   if just_one:
      check_ID = case_ID.upper()
      if check_ID == 'FIRST':
         ID_order = ID_order[:1]    # Select first test case in the file
      else:
         if not case_ID.isdigit() or len(case_ID) > 3:
            (directory, filename) = os.path.split(fileid)
            showerror('ERROR!',
                      'Directory  ' + directory + '\n\n' +
                      'Filename:  ' + filename + '\n\n' +
                      '"' + case_ID + '" does not have a valid test case ID format.')
            return 'Invalid case ID'
         key = case_ID.strip().rjust(3,'0')
         if key not in ID_order:
            (directory, filename) = os.path.split(fileid)
            showerror('ERROR!',
                      'Directory  ' + directory + '\n\n' +
                      'Filename:  ' + filename + '\n\n' +
                      '"' + case_ID + '" is not a test case ID in this file.\n\n' +
                      'Valid IDs are: ' + str(ID_order))
            return 'Unrecognized case ID'
         else:
            ID_order = [key]


   # For grid mode, the number of cases that can be displayed depends on
   # screen size and user input ("rows =").  The grid_mode_setup
   # function is called to calculate cell height and width as well as
   # the number of rows and columns in the display and the number of
   # cases that can be displayed.  It then sets the main window's
   # geometry and creates a backing frame in the main window for the
   # grid cells.
   else:
      (backing, cell_height, cell_width, max_rows, nbr_cases
      ) = grid_mode_setup(ID_order, file_rows, reduce_size)

      # Now that the displayable number of cases has been updated in
      # nbr_cases, the ID_order list can be trimmed to contain only the
      # IDs that can be displayed.
      ID_order = ID_order[:nbr_cases]


   # ===================================================================
   # Loop thru the cases dictionary and put the packed contents of each
   # test case into its own pack_holder frame.  For grid mode the
   # pack_holder frames are fixed-size grid cells.  For "just one" mode
   # the pack_holder frame is an expandable Toplevel window frame.

   # Reinitialize these global values to create the current display.
   totcells = 0         # Running total of the number of cells processed
   row = 0              # Row coordinate of grid cell
   col = 0              # Column coordinate of gridcell
   comment_labels = []  # List of comment labels in the display
   first_comment = True # Used for resizing

   # VVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVV
   # Start of label packing loop.

   for ID in ID_order:

      # For "just one" mode, build a pack_holder frame attached to a
      # Toplevel window to hold the packing results of the single test.
      # Even though this is a loop thru the ID_order list, only one
      # frame is built.  The above "ID_order = [key]" for "just one"
      # mode put only one dictionary key in ID_order.  Thus the loop
      # makes only one pass.
      if just_one:
         jo_win = just_one_mode_setup() # Get customized Toplevel window
         pack_holder = Frame(jo_win, bd=1, relief=SOLID)
         pack_holder.pack(expand=YES, fill=BOTH)

      # For grid mode, for the grid's frame use the backing frame that
      # was created earlier when the grid's Toplevel window was created
      # and customized.  Insert a grid cell (a pack_holder) into the
      # backing frame for each pass thru the loop.
      else:
         # Insert a fixed-size frame into the backing frame.  This is
         # the grid cell for this test case.
         pack_holder = Frame(backing, bd=1, relief=SOLID,
                           height=cell_height, width=cell_width)
         # Place the cell at a particular row and column in the grid.
         # Provide padding around the cell to let the backing frame's
         # color be visible around the cell's edges.
         pack_holder.grid(row=row, column=col, padx=gapx, pady=gapy, sticky=NSEW)

         # Make grid resizeable.
         backing.rowconfigure(row, weight=1)
         backing.columnconfigure(col, weight=1)

         # This propagate call keeps the cell frame (the parent) at the
         # specified "fixed" size relative to its contents.  Otherwise
         # it would shrink/expand as things were packed into it. All
         # the grid cells are the same size, but that size may grow or
         # shrink as the whole Toplevel window grows or shrinks by a
         # mouse drag.
         pack_holder.pack_propagate(0)

      # ----------------------------------------------------------------
      # From here on, grid mode and "just one mode" operate the same.
      # They fill in the pack_holder frame that has been built for each
      # mode.  For "just one" mode, there is only one pack_holder frame
      # and one test case.  For grid mode, a new pack_holder frame will
      # be built for each pass thru the loop.  Each such frame will be a
      # grid cell to hold one test case's results.  Passes thru the loop
      # will continue until all the test cases with an ID in ID_order
      # have been displayed.

      # ID is a 3-character string.  It was derived from the first field
      # in the file line for this set of label packing instructions.  It
      # was then padded on the left with 0's.  Here the leading 0's are
      # stripped to provide a neater display, with '000'->'0'
      ID_lab = ID.lstrip('0')
      if ID_lab == '': ID_lab='0'

      # Put a case ID label at the top of the cell.
      case_ID = Label(pack_holder, text='CASE ' + ID_lab, bd=1, relief=SOLID,
                      bg=bg_case_ID, fg=fg_case_ID, font=ft_case_ID)
      case_ID.pack(side=TOP, anchor=CENTER, fill=X)

      # Given the construction of the ID_order list in parallel with the
      # construction of the cases dictionary, this is now a "Should Not
      # Occur" (but just in case).
      if ID not in cases:
         missing = Label(pack_holder, text='THIS CASE WAS NOT FOUND',
                       bg=bg_error, fg=fg_error, font=ft_error)
         missing.pack(expand=YES, fill=BOTH)

         if not just_one:          # Calculate coordinates for next cell
            (totcells, row, col) = advance(totcells, row, col, max_rows)
         continue

      # ----------------------------------------------------------------
      # In the cases dictionary the ID key's value is a list of one or
      # more label packing strings.  Loop thru the label packing strings
      # for the test case identified by the ID key.  Color each label
      # and pack it into the test case's assigned cell frame (grid mode)
      # or into the expandable window frame ("just one" mode).  The
      # "pack_holder" variable is used for whichever parent frame is
      # appropriate for the current mode.

      case_length = len(cases[ID]) # Number of labels to pack in this cell
      label_nbr = 0     # Order of packing number displayed in the label
      for i in range(case_length):
         org_lab = cases[ID][i]  # Get a packing string from the list
         if org_lab.strip() == 'IGNORE': continue

         # Handle building and packing an information "TEXT=" label.
         if org_lab[:5] == 'TEXT=':
            info_text = org_lab[5:]

            # This is the instantiation of the "TEXT=" label.
            wrap = info_label_width - 30           # Initial wrap length
            comment_label = Label(pack_holder, text=info_text, padx=5,
                          bg=bg_text, fg=fg_text, font=ft_text, justify=LEFT,
                          wraplength=wrap, anchor=W, bd=10, relief=RIDGE)
            comment_label.pack(side=BOTTOM, expand=NO, fill=X)

            # Add this label to the list of all comment labels in the
            # display.  This list is used to change the wrap length for
            # all comment labels on a window resize.  The bind to the
            # onResize event handler is done for only the first label in
            # the list -- all comment labels change when it changes.
            comment_labels.append(comment_label)
            if first_comment:
               comment_label.bind('<Configure>', onResize)
               first_comment = False

         # Handle building and packing any other type of label.
         else:
            # Get the color for each label; each side has its own color.
            text_lab = org_lab.replace(' ','')
            text_lab = text_lab.replace(',',', ')
            if   "side=RIGHT"  in text_lab: bg_lab = bg_RIGHT
            elif "side=LEFT"   in text_lab: bg_lab = bg_LEFT
            elif "side=BOTTOM" in text_lab: bg_lab = bg_BOTTOM
            else:                           bg_lab = bg_TOP

            # The text in the first label packed is colored specially.
            if i == 0:
               fg_lab = fg_1
               ft_lab = ft_1
            else:
               fg_lab = fg_a
               ft_lab = ft_a

            # Save unabbreviated pack string for eval to do the actual pack.
            eval_lab = text_lab

            # Abbreviate some options and values in the label text of the
            # pack string if the abbreviate option was selected.
            if abbreviate:
               text_lab = text_lab.replace('side=',   ' side=')
               text_lab = text_lab.replace('fill=',   ' fill=')
               text_lab = text_lab.replace('expand=', ' exp=')
               text_lab = text_lab.replace('anchor=', ' anc=')
               text_lab = text_lab.replace('=NONE',   '=*')
               text_lab = text_lab.replace('=BOTH',   '=B')
               text_lab = text_lab.replace('=TOP',    '=T')
               text_lab = text_lab.replace('=BOTTOM', '=B')
               text_lab = text_lab.replace('=LEFT',   '=L')
               text_lab = text_lab.replace('=RIGHT',  '=R')
               text_lab = text_lab.replace('=YES',    '=Y')
               text_lab = text_lab.replace('=NO',     '=N')
               text_lab = text_lab.replace('=*',      '=N')

            # This part of the label text displays the packing order.
            label_nbr += 1        # Bump the count to show in this label
            nbr_lab = '#' + str(label_nbr)

            # This is the instantiation of a non-"TEXT=" label.
            label = Label(pack_holder, text=nbr_lab + '  ' + text_lab, padx=5,
                          bg=bg_lab, fg=fg_lab, font=ft_lab, bd=1, relief=SOLID)

            # Finally, pack the label into the cell's frame.
            try:
               # This does the pack of the label, using the expression
               # dynamically created above.  The options have been
               # checked earlier to a large degree, so this should
               # usually not fail.  One way to cause a failure is to
               # have duplicate options for a label, e.g., a side=TOP
               # occurring twice in the pack string; Pack Grid does not
               # try to catch this, but instead lets tkinter report the
               # error in the "except" below.
               eval('label.pack(' + eval_lab + ')')
            except SyntaxError as exdata:
               bad = Label(pack_holder,
                           text='FAILURE FOR CASE ID: ' + ID_lab + '\n\n' + str(exdata),
                           bg=bg_error, fg=fg_error, font=ft_error)
               bad.pack(expand=YES, fill=BOTH)

            # End of the else processing for this label (a non-"text=" label)

         # Now go back to the top of the "for i in range(case_length)"
         # loop to get the next pack string and start a new label.


      # Exit from the "for i in range(case_length)" loop means there are
      # no # more pack strings for this ID, so this test case is now
      # finished.

      # For grid mode, get ready for the next test case by calculating
      # the coordinates for the next cell which will be used for that
      # test case.
      if not just_one:
         (totcells, row, col) = advance(totcells, row, col, max_rows)

      # For grid mode, now go pack to the top of the loop to process the
      # next test case.  When there are no more test cases to display,
      # exit the loop and return to pack_grid.pyw.

      # For "just one" mode, this completes the one and only pass thru
      # the loop to display the single selected test case.  Exit the
      # loop and return to pack_grid.pyw.

   # After exit from the "for ID in ID_order" loop
   # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

   return


# ======================================================================
# This function is called by build_display_windows for grid mode.
# Create a Toplevel window and set up its geometry.  Also set up the
# grid cell width and height, and the number of rows in the display.  If
# needed, the display can spill to a second column.  The desired number
# of rows can be specified in a "rows =" line in the file.  For "rows =
# max" (the default), the maximum number of rows is calculated based on
# the number of test cases, but with a minimum cell height as a
# constraint.  The cells are stretched or shrunk vertically to fill the
# screen (or half-screen for "reduced size"), but within the minimum
# cell height constraint.

def grid_mode_setup(ID_order, file_rows, reduce_size):
   global gapx, gapy, info_label_width

   grid_window = Toplevel()
   grid_window.title(fileid)
   grid_window.protocol('WM_DELETE_WINDOW',grid_window.destroy)
   def onDoubleLeftClick(event): grid_window.destroy()
   grid_window.bind('<Double-1>', onDoubleLeftClick)

   # Get tkinter's version of the screen size, which is NOT the actual
   # screen size when measured in pixels.
   org_scr_width  = grid_window.winfo_screenwidth()
   org_scr_height = grid_window.winfo_screenheight()

   # A small width adjustment (14) is done below since tkinter's
   # "pixels" make it impossible to place the display exactly on the
   # left side of the physical screen -- an offset of 0 does not produce
   # a true 0 offset.  This tweak will slightly inset the display on
   # both sides to make it appear these side gaps are intentional.  This
   # workaround is probably wrong on some physical screens.
   scr_width  = org_scr_width  - 14

   # A height adjustment is made to avoid having part of the display
   # disappear under the Windows task bar.  This workaround is probably
   # wrong on some physical screens.
   scr_height = org_scr_height - 65

   # For the "reduce" option, reduce the size of the displayed window
   # by about 1/2 and then center it on the screen.  Do this by sizing
   # the window originally as if the screen were half its true size.
   # Then insert that window into the full-size screen.  The same
   # minimum cell-height constraint exists as for the full-size screen.
   if reduce_size:
      full_scr_width = scr_width    # Used later for "reduce"d display
      full_scr_height = scr_height  #   to insert into full-size screen

      scr_width =  scr_width  // 2  # Displays better on large screens
      scr_height = scr_height // 2  #

   cell_width = (scr_width  - 4*gapx) // 2         # Cell width for grid
   info_label_width = cell_width  # Save for wrapping text in info label

   # This calculates the maximum number of rows possible on this screen,
   # given the screen's height and the minimum allowed cell height plus
   # gaps for backing frame visibility.
   min_cell_height = 76               # CONSTRAINT:  Minimum cell height
   max_rows_physical = scr_height // (min_cell_height + 2*gapy)

   nbr_cases = len(ID_order)          # Number of test cases in the file

   # -------------------------------------------------------------------
   # If "rows=max", either by default of by user entry in the file, the
   # actual number of rows was not specified, so it must be calculated.
   # The cases will be displayed in two "equal" columns, unless only one
   # case is to be displayed.
   if file_rows == 'MAX':
      file_rows_int = max_rows_physical

      # Limit number of cases to what will fit in two columns.
      max_cells_in_two_cols = 2 * file_rows_int
      org_nbr_cases = nbr_cases
      if nbr_cases > max_cells_in_two_cols:
         nbr_cases = max_cells_in_two_cols
         grid_window.withdraw()           # Hide window to avoid clutter
         (directory, filename) = os.path.split(fileid)
         showwarning('WARNING!',
                     'Directory  ' + directory + '\n\n' +
                     'Filename:  ' + filename + '\n\n' +
                     'There are ' + str(org_nbr_cases) + ' test cases, ' +
                     'which will not all fit in two columns.  Only ' +
                     str(max_cells_in_two_cols) + ' will be displayed .')
         grid_window.state(NORMAL)                  # Restore the window
      # Output: (reduced) nbr_cases

      # Using (reduced) number of cases, distribute the cases as evenly
      # as possible over two columns.  Calculate the number of rows
      # required.  If nbr_cases=1, use a single column.
      half_nbr_cases = int(max((nbr_cases/2 + 0.5) // 1, 1))  # Round up
      cell_height = scr_height // half_nbr_cases - 2*gapy
      file_rows_int = half_nbr_cases
      file_rows = str(file_rows_int)       # Restore string for messages
      # Output: (reduced) number of cases, file_rows, cell_height

   # -------------------------------------------------------------------
   # The user specified the desired number of rows via "rows = nbr".
   # Use that number of rows if possible.  Otherwise use the maximum
   # possible rows for this screen.
   else:
      org_file_rows = file_rows
      file_rows_int = int(file_rows)

      # Limit file rows to the number of file test cases.
      file_rows_int = min(file_rows_int, nbr_cases)

      # Limit file rows to the number of rows physically possible.
      if file_rows_int > max_rows_physical:
         file_rows_int = max_rows_physical
         grid_window.withdraw()           # Hide window to avoid clutter
         (directory, filename) = os.path.split(fileid)
         showwarning('WARNING!',
                     'Directory  ' + directory + '\n\n' +
                     'Filename:  ' + filename + '\n\n' +
                     'The "rows =" value (' + org_file_rows + ') ' +
                     'listed in this file is too large for this screen. ' +
                     'Instead, the largest possible number of rows ' +
                     '(' + str(max_rows_physical) + ') will be used.')
         grid_window.state(NORMAL)                  # Restore the window

      # Limit number of cases to what will fit in two columns using the
      # (reduced) number of user-specified rows.
      max_cells_in_two_cols = 2 * file_rows_int
      org_nbr_cases = nbr_cases
      if nbr_cases > max_cells_in_two_cols:
         nbr_cases = max_cells_in_two_cols
         grid_window.withdraw()           # Hide window to avoid clutter
         (directory, filename) = os.path.split(fileid)
         showwarning('WARNING!',
                     'Directory  ' + directory + '\n\n' +
                     'Filename:  ' + filename + '\n\n' +
                     'There are ' + str(org_nbr_cases) + ' test cases, ' +
                     'which will not all fit in two columns.  Only ' +
                     str(max_cells_in_two_cols) + ' will be displayed .')
         grid_window.state(NORMAL)                  # Restore the window

      # Using (reduced) number of cases and (reduced) number of rows
      # the user specified, calculate the cell height.
      cell_height = scr_height // file_rows_int - 2*gapy
      if nbr_cases == 1: cell_height = scr_height - 4*gapy
      file_rows = str(file_rows_int)       # Restore string for messages
      # Output: (reduced) number of cases, file_rows, cell_height


   # -------------------------------------------------------------------
   # Finish window size and placement calculations and set the geometry.

   max_rows = file_rows_int

   if nbr_cases/max_rows > 1:
      disp_width = scr_width          # Multi column
   else:
      disp_width = scr_width // 2     # Single column
   disp_height = scr_height           # Same height for single and multi

   if reduce_size:
      # Offsets are larger when inserting a reduced-size window into a
      # full-sized screen.  Here disp_* are half the non-reduced sizes.
      x_offset = (full_scr_width  - disp_width)  // 2
      y_offset = (full_scr_height - disp_height) // 2
   else:
      # These are the offsets for inserting a non-reduced-size window.
      # The scr_* values are the full_scr_* values (i.e., not halved).
      x_offset = (scr_width  - disp_width)  // 2
      y_offset = (scr_height - disp_height) // 2

   # Set up the grid window geometry.
   grid_geom = (str(disp_width) + 'x' + str(disp_height) + '+' +
               str(x_offset)   + '+' + str(y_offset))

   grid_window.focus_force()       # If focus_force is done AFTER the
                                   #   geometry set, the geometry set
                                   #   does not work properly (?)
   grid_window.geometry(grid_geom)

   # Set up a colored backing frame to show color between the cell
   # frame edges.
   backing = Frame(grid_window, bg='blue')
   backing.pack(expand=YES, fill=BOTH)

   return (backing, cell_height, cell_width, max_rows, nbr_cases)


# ======================================================================
# This function is called by build_display_windows after the current
# test case has been processed.  It adjusts the row and column grid
# coordinates to get to the cell for the next test case.

def advance(totcells, row, col, max_rows):
   totcells += 1     # Late add; result is number of cell just processed
   if totcells % max_rows == 0:
      col += 1       # Start a new column
      row = 0        # Go to the top of the new column
   else:
      row += 1       # Stay in same column and go to next row in it
   return (totcells, row, col)


# ======================================================================
# This function is called by build_display_windows to build a customized
# Toplevel window for "just one" mode.

def just_one_mode_setup():
   global info_label_width
   jo_win = Toplevel()
   jo_win.title(fileid)
   jo_win.protocol('WM_DELETE_WINDOW',jo_win.destroy)
   scr_width  = jo_win.winfo_screenwidth()
   scr_height = jo_win.winfo_screenheight()
   info_label_width = scr_width // 2  # Save for text wrap in info label
   jo_width =  str(scr_width  // 2)
   jo_height = str(scr_height // 2)
   jo_x_off =  str(scr_width  // 4)
   jo_y_off =  str(scr_height // 4)
   geom = jo_width + 'x' + jo_height + '+' + jo_x_off + '+' + jo_y_off
   jo_win.geometry(geom)

   def onDoubleLeftClick(event):
      jo_win.destroy()
   jo_win.bind('<Double-1>', onDoubleLeftClick)

   jo_win.focus_force()
   return jo_win


# ======================================================================
# This function is called when a window resize occurs.  It sets a new
# text wrap length for each comment ("text=") label in the display (if
# there are any) because the resize probably changes the size of those
# labels.  The comment labels were put in a list when the original
# display was built.  This event handler does the wrap update for each
# label in the list although the handler was bound for only the first
# comment label that was built.

def onResize(event):
   for comment_label in comment_labels:
      comment_label.configure(wraplength=event.width - 30)
